'''
This module can help users quickly modify the default special
key identifiers.
'''

import_id = '__import__'
imported_targets_id = '__import__'
args_id = '__args__'
kwargs_id = '__kwargs__'
run_id = '__run__'